# Disegno il grafico regime/velocità partendo da velocità zero fino a velocità
# massima.
# Il conto è eseguito su un grafico normalizzato, ovvero sia il regime massimo
# che la velocità massima sono pari a 1.
# L'ottimizzazione si basa sul massimizzare l'area sottesa al grafico sopradet-
# to. Massimizzando quest'area si massimizza il regime medio durante l'accelera-
# zione, e di conseguenza la potenza e quindi l'accelerazione.
# Non si tiene conto della resistenza aerodinamica, ma comunque non avrebbe ef-
# fetto sul risultato finale.

clear;
clc;
close all;

# Inserimento dati
v1 = 0;            # Vel normalizzata in prima marcia. Se v1 = 0 allora anche la
                   # prima marcia viene ottimizzata.
nr = 6;            # Numero di rapporti
np = 10000;        # Numero individui
ng = 500;          # Numero generazioni
nf = 4;            # Numero individui per torneo

# Popolazione iniziale
if v1 != 0
  R = [ones(np,1)*v1, sort(v1+rand(np, nr-2)*(1-v1), 2), ones(np,1)];
else
  R = [sort(rand(np, nr-1), 2), ones(np,1)];
endif
fit = zeros(np, ng);
Rtot = zeros(np,nr,ng);
for i = 1:ng
  Rtot(:,:,i) = R;
  # Calcolo il fit degli individui
  fit(:,i) = sum(R, 2) - sum(R(:,1:end-1).^2./R(:,2:end), 2);
  
  # Seleziono gli individui per la riproduzione
  sel = randi(np, np, nf);
  vin = max(fit(:,i)(sel), [], 2);
  ind = (fit(:,i)'==vin).*linspace(1,np,np);
  ind(ind==0) = [];
  
  # Applico una mutazione casuale e ricreo la matrice della generazione succes-
  # siva
  if v1 != 0
    R = [ones(np,1)*v1, R(ind,2:end-1).*(1+randn(np, nr-2)*0.05), ones(np,1)];
  else
    R = [R(ind,1:end-1).*(1+randn(np, nr-1)*0.05), ones(np,1)];
  endif
  
  # Plotto valor medio e scarto quadratico del fit
  media = mean(fit(:,1:i));
  dev = std(fit(:,1:i));
  h1 = figure(1);
  errorbar(media, dev);
  xlabel("Generazioni", 'FontWeight', 'bold', 'Fontsize', 14);
  ylabel("Media e dev. standard", 'FontWeight', 'bold', 'Fontsize', 14);
  title("Fit", 'FontWeight', 'bold', 'Fontsize', 14);
  grid on;
endfor

# Trovo l'individuo con il fit migliore
best = max(max(fit));
ind = find(fit==best);
gen = ceil(ind/np);
diary on;
disp(["Individuo con il fit migliore nella generazione ", num2str(gen), " su ", num2str(ng)]);

# Velocità marce individuo migliore
vel = Rtot(ind-(gen-1)*np,:,gen);
disp(["Velocità massima normalizzata in ogni marcia:"]);
disp(vel);
diary off;

# Plotto il grafico con le velocità in ogni marcia
x = [0, reshape([vel;vel], 1, nr*2)(1:end-1)];
y = [0, reshape([ones(1, nr-1); (vel(1:end-1)./vel(2:end))], 1, (nr-1)*2), vel(end)];
h2 = figure(2);
plot(x, y);
xlabel("Velocita", 'FontWeight', 'bold', 'Fontsize', 14);
ylabel("Regime", 'FontWeight', 'bold', 'Fontsize', 14);
title("Velocita/Regime normalizzato", 'FontWeight', 'bold', 'Fontsize', 14);
grid on;
print(h1, "generazioni.pdf");
print(h2, "rapportatura.pdf");