clear;
clc;
t_ini = time();
## Ottimizzazione genetica dei rapporti del cambio, posizione del baricentro
## e regime di cambio marcia in ogni rapporto.
## Viene trovata la rapportatura ottimale per un dato veicolo in un dato
## rettilineo, con determinate velocità iniziale e finale.
## Per questo calcolo il file di dati deve rappresentare un solo veicolo,
## quindi i vari paramentri (come posizione e altezza del baricentro eccetera)
## devono essere valori singoli e non matrici.
## Le opzioni modificabili sono:
## - individui della popolazione: Dimensione iniziale della popolazione, più è
##   grande e più probabilità c'è di trovare la soluzione corretta
## - dimensione del torneo: numero di individui scelti random tra la popolazione
##   che si sfideranno. Solo il più forte di questi genererà la popolazione
##   successiva, e vengono fatti tanti tornei quanto è il numero degli
##   individui della popolazione, per tornare alla dimensione iniziale. Non
##   c'è un valore corretto da usare, ma più questo numero è grande, più è
##   probabile che siano sempre gli stessi a riprodursi.
##   Lunghezza iniziale del vettore risultati: è solo una stima, se non è
##   sufficiente viene allungato durante il ciclo.

## OPZIONI
np = 500;         ## Individui della popolazione. Deve essere un numero pari
dk = 4;           ## Dimensione del torneo: quanti individui si sfidano ogni volta
li = 200;         ## Lunghezza iniziale vettore risultati. Corrisponde al numero di
                  ## generazioni che si stima di avere
nr = 6;           ## Numero di rapporti
rmax = 20;        ## Valore massimo del rapporto
hmax = 1;         ## Altezza max baricentro
hmin = 0.3;       ## Altezza min baricentro
bmax = 2.5;       ## Posizione "b" max baricentro
bmin = 0;         ## Posizione "b" min baricentro
auto_TA;
##______________________________________________________________________________
##
##                INIZIO CALCOLO
##______________________________________________________________________________
diary on;
disp(ctime(t_ini));
disp(["OTTIMIZZAZIONE RAPPORTATURA DEL CAMBIO, POSIZIONE DEL BARICENTRO"])
disp(["E REGIME CAMBIO MARCIA"])
disp("")
disp("Un veicolo viene fatto accelerare e quindi frenare in un rettilineo di")
disp("lunghezza data, con velocità iniziale e finale date.")
disp("")
disp("Dati veicolo:")
disp(["Massa totale: ..................................", num2str(m), "kg"])
disp(["Passo: .........................................", num2str(p*1000), "mm"])
disp(["Ripartizione pesi front/post: ..................", num2str(b/p*100), "/", num2str((p-b)/p*100), "%"])
disp(["Altezza baricentro: ............................", num2str(h*1000), "mm"])
disp(["Raggio ruota posteriore: .......................", num2str(Rr*1000), "mm"])
disp(["Raggio ruota anteriore: ........................", num2str(Rf*1000), "mm"])
disp(["Inerzia ruota posteriore: ......................", num2str(I_r), "Kg*m^2"])
disp(["Inerzia ruota anteriore: .......................", num2str(I_f), "Kg*m^2"])
disp(["Inerzia motore ridotta all'albero: .............", num2str(I_m), "Kg*m^2"])
disp(["Inerzia albero primario cambio: ................", num2str(I_p), "Kg*m^2"])
disp(["Inerzia albero secondario cambio: ..............", num2str(I_s), "Kg*m^2"])
disp(["Rapporto primario: .............................", num2str(R_p), ":1"])
disp(["Rapporto finale: ...............................", num2str(R_f), ":1"])
disp(["Area frontale: .................................", num2str(Af), "m^2"])
disp(["Coefficiente di resistenza aerodinamica: .......", num2str(Cd)])
disp(["Coefficiente di portanza aerodinamica: .........", num2str(Cl)])
disp(["Coefficiente di momento aerodinamico: ..........", num2str(Cm)])
disp(["Coefficiente di attrito gomma/asfalto: .........", num2str(mu)])
disp(["Coefficiente di resistenza al rotolamento: .....", num2str(mur)])
disp(["Altezza massima del baricentro: ................", num2str(hmax*1000), "mm"])
disp(["Altezza minima del baricentro: .................", num2str(hmin*1000), "mm"])
disp(["Posizione \"b\" massima del baricentro: ..........", num2str(bmax*1000), "mm"])
disp(["Posizione \"b\" minima del baricentro: ...........", num2str(bmin*1000), "mm"])
disp(["Massimo rapporto: ..............................", num2str(rmax)])
if tr == "r"
  trr = "Posteriore";
elseif tr == "f"
  trr = "Anteriore";
endif
if fdu == "fr"
  FDU = "Anteriore e posteriore";
elseif fdu == "f"
  FDU = "Anteriore";
endif
disp(["Tipo di trazione: ..............................", trr])
disp(["Freno utilizzato: ..............................", FDU])
disp("")
disp("Dati motore:")
disp(" rpm     Coppia [Nm] Potenza [CV]")
rpm_disp = round(CC(1,:))';
coppia_disp = (round(CC(2,:)*10)/10)';
potenza_disp = (round((CC(1,:)*pi/30).*(CC(2,:))/735*10)/10)';
## Creo gli spazi tra i numeri in tabella
sp = [];
sp = num2str(sp);
for kk = 1:size(CC, 2)
  sp = [sp; "       "];
endfor
disp([num2str(rpm_disp), sp, num2str(coppia_disp), sp, num2str(potenza_disp)])
disp("")
disp("Dati Percorso:")
disp(["Lunghezza rettilineo: ..........................", num2str(Lr), "m"])
disp(["Suddivisione rettilineo: .......................", num2str(ns), " tratti"])
disp(["Pendenza strada: ...............................", num2str(tan(alpha)*100), "%"])
disp(["Velocità iniziale: .............................", num2str(round(vi*3.6*10)/10), "km/h"])
disp(["Velocità finale: ...............................", num2str(round(vf*3.6*10)/10), "km/h"])

## CREO LA POPOLAZIONE INIZIALE
## Ogni riga della matrice sarà un individuo, ogni colonna un rapporto
## da 1 a nr. Le due colonne dopo i rapporti saranno, rispettivamente, posizione
## longitudinale "b" e altezza "h" del baricentro. Le ultime sei colonne
## saranno i regimi di cambio marcia per ogni rapporto.
Rc = [sort(rand(np,nr)*rmax, 2, "descend"), bmin+rand(np, 1)*(bmax-bmin), hmin+rand(np, 1)*(hmax-hmin), CC(1,1)+rand(np, nr)*(CC(1,end)-CC(1,1))];

## INIZIO ALGORITMO
disp("")
disp("INIZIO CALCOLO OTTIMIZZAZIONE GENETICA");

disp(["Popolazione: ...................................", num2str(np), " individui"]);
disp(["Dimensione del torneo di selezione: ............", num2str(dk), " individui"]);
disp(["Stima generazioni necessarie: ..................", num2str(li)]);
disp("");
disp("Per fermare la simulazione, aprire il file choose.txt e digitare qualunque")
disp("lettera al posto di n.")
disp("")
disp("Generazione | Tempo elemento migliore | Media tempi popolazione | deviazione standard")

## Inizializzo vettori. Li creo di una lunghezza casuale, non conoscendo la
## durata del ciclo
r_tot = zeros(np, nr*2+2, li);
t_tot = zeros(np, li);
b_bestg = zeros(1, li);
h_bestg = zeros(1, li);

## Valori iniziali per il ciclo
choose = "n";
save choose.txt choose;
k = 1;
while choose == "n"
  ## CALCOLO IL FITNESS DI OGNI INDIVIDUO
  ## In questo caso è rappresentato dal tempo di percorrenza del rettilineo.
  ## Minore il tempo, migliore il fitness
  fit = zeros(1,np);
  for i = 1:np
    [tt, xx, vv, rpmm, ccm, ccr, pgas, Nr, Nf] = rett(m, p, Rc(i,nr+1), Rc(i,nr+2), Rr, Rf, I_r, I_f, I_m, I_p, I_s, Rc(i,1:nr), R_p, R_f, CC, Rc(i,end-nr+1:end), Af, Cd, Cl, Cm, rho, mu, mur, alpha, Lr, vi, vf, ns, fdu, tr);
    fit(i) = tt(1,1,1,1,1,1,end);
  endfor
  
  ## SALVO I DATI UTILI
  ## Estendo la matrice risultati se ho superato il numero di generazioni stimate
  if k != 1
    if (k-1)/li-floor((k-1)/li) == 0
      r_tot1 = zeros(np, nr*2+2, ((k-1)/li+1)*li);
      r_tot1(:,:,1:size(r_tot, 3)) = r_tot;
      r_tot = r_tot1;
      t_tot = [t_tot, zeros(np, li)];
      b_bestg = [b_bestg, zeros(1,li)];
      h_bestg = [h_bestg, zeros(1,li)];
      disp(["Raggiunto il numero stimato di generazioni, estendo la matrice risultati"]);
      disp(["Nuova lunghezza: ", num2str(size(t_tot, 2))]);
    endif
  endif
  r_tot(:,:,k) = Rc;
  t_tot(:,k) = fit';

  ## SELEZIONO GLI INDIVIDUI PER LA RIPRODUZIONE
  ## Viene utilizzato il metodo del torneo: vengono scelti, random, un numero
  ## "dk" di individui che si dovranno sfidare. Ad ogni torneo viene tenuto solo
  ## l'individuo vincente. Si fanno tanti tornei quanti ne servono per tornare
  ## al numero della popolazione iniziale.
  
  tor = randi(np, np, dk);              ## Indice degli individui che si sfideranno
  FIT = zeros(np, dk);                  ## Tempi degli individui che si sfideranno
  for j = 1:np
    FIT(j,:) = fit(tor(j,:));
  endfor
  tch = min(FIT, [], 2);                ## Tempi degli individui vincenti
  it = ones(np, np).*fit-tch;
  it(it!=0) = 1;
  it = not(it).*linspace(1,np,np);
  it(it == 0) = [];                     ## Indice degli individui che si riprodurranno

  ## CREO LA SECONDA GENERAZIONE
  ## Scelgo per la riproduzione gli individui degli indici di "it"
  ## Applico una piccola mutazione random a questi individui, poi inverto le
  ## prime due righe di metà matrice, in modo che i figli abbiano metà tratti
  ## di un genitore, metà dell'altro.

  ## Creo il vettore Rc avente solo i vincitori del torneo, con una piccola
  ## mutazione random. Se alcuni valori superano i limiti, li riporto entro
  ## i limiti. Non modifico il valore del rapporto massimo, perchè più che un
  ## limite, è un numero di riferimento da cui partire.
  Rc = Rc(it, :).*(1+randn(np, nr*2+2)*0.025);
  Rce = Rc(:,end-nr);               ## Tengo la colonna di altezze baricentro
  Rcr = Rc(:,nr+3+floor(nr/2):end); ## Tengo le ultime colonne dei regimi cambio marcia
  Rce(Rce > hmax) = hmax;
  Rce(Rce < hmin) = hmin;
  Rcr(Rcr > CC(1,end)) = CC(1,end);
  Rcr(Rcr < CC(1,1)) = CC(1,1);
  
  ## Applico il crossover: ogni 2 genitori, scambio la parte iniziale dei
  ## rapporti e del regime di cambio marcia, e la dimensione "b"
  Rcn = zeros(np, floor(nr/2)*2+1);
  for i = 1:2:np
    Rcn(i:i+1,:) = [flipud(Rc(i:i+1,1:floor(nr/2))), flipud(Rc(i:i+1,nr+1)), flipud(Rc(i:i+1,nr+3:nr+2+floor(nr/2)))];
  endfor
  Rcr1 = Rcn(:,floor(nr/2)+2:end);
  Rcr1(Rcr1 > CC(1,end)) = CC(1,end);
  Rcb = Rcn(:,floor(nr/2)+1);
  Rcb(Rcb > bmax) = bmax;
  Rcb(Rcb < bmin) = bmin;
  Rc = [Rcn(:,1:floor(nr/2)), Rc(:,floor(nr/2)+1:nr)];
  Rc = [sort(Rc, 2, "descend"), Rcb, Rce, Rcr1, Rcr];
  
  ## CALCOLO LE STATISTICHE DI ACCELERAZIONE PER OGNI GENERAZIONE
  t_mean = mean(t_tot)(:,1:k);
  t_std = std(t_tot)(:,1:k);
  disp(["     ", num2str(k), "                 ", num2str(min(t_tot(:,k))), "                 ", num2str(t_mean(end)), "                 ", num2str(t_std(end))]);

  ## plotto l'andamento delle statistiche attraverso le generazioni
  figure(1)
  errorbar(t_mean, t_std);
  xlabel("Generazioni", 'FontWeight', 'bold', 'Fontsize', 14);
  ylabel("Media e dev. standard", 'FontWeight', 'bold', 'Fontsize', 14);
  title("Tempi accelerazione popolazione", 'FontWeight', 'bold', 'Fontsize', 14);
  grid on;
  
  ## plotto l'andamento della posizione longitudinale del baricentro
  ##normalizzato rispetto al passo dell'individuo migliore di ogni generazione
  figure(2)
  ind_best = find(t_tot(:,k)==min(t_tot(:,k)))(1);
  b_bestg(k) = r_tot(ind_best, nr+1, k)/p;
  plot(b_bestg(1:k));
  xlabel("Generazioni", 'FontWeight', 'bold', 'Fontsize', 14);
  ylabel("\"b\" normalizzata", 'FontWeight', 'bold', 'Fontsize', 14);
  title("Posizione longitudinale baricentro", 'FontWeight', 'bold', 'Fontsize', 14);
  grid on;
  
  ## plotto l'andamento dell'altezza del baricentro
  ## normalizzato rispetto al passo dell'individuo migliore di ogni generazione
  figure(3)
  h_bestg(k) = r_tot(ind_best, nr+2, k)/p;
  plot(h_bestg(1:k));
  xlabel("Generazioni", 'FontWeight', 'bold', 'Fontsize', 14);
  ylabel("\"h\" normalizzata", 'FontWeight', 'bold', 'Fontsize', 14);
  title("Altezza del baricentro", 'FontWeight', 'bold', 'Fontsize', 14);
  grid on;
  
  ## Posizione x-y  normalizzata del baricentro
  figure(4)
  plot(b_bestg(1:k), h_bestg(1:k));
  axis([0 1 0 1]);
  xlabel("Posizione longitudinale normalizzata", 'FontWeight', 'bold', 'Fontsize', 14);
  ylabel("Posizione verticale normalizzata", 'FontWeight', 'bold', 'Fontsize', 14);
  title("Posizione del baricentro", 'FontWeight', 'bold', 'Fontsize', 14);
  grid on;
  
  ## Valori per il ciclo successivo
  k = k+1;
  load choose.txt
  save r_tot.txt r_tot
  save t_tot.txt t_tot
  save b_bestg.txt b_bestg
  save h_bestg.txt h_bestg
endwhile

## Trovo le caratteristiche degli individui migliore/peggiore nell'ultima generazione
t_tot(:,k:end) = [];
r_tot(:,:,k:end) = [];
t_best = min(t_tot(:,end));
t_worst = max(t_tot(:,end));
ind_best = find(t_tot(:,end)==t_best)(1);
ind_worst = find(t_tot(:,end)==t_worst)(1);
b_best = r_tot(:,:,end)(ind_best,nr+1);
h_best = r_tot(:,:,end)(ind_best,nr+2);
disp("")
disp("Caratteristiche ultima generazione")
disp(["Tempo percorrenza individuo migliore: ..........", num2str(t_best), "s"])
disp(["Tempo percorrenza individuo peggiore: ..........", num2str(t_worst), "s"])
disp(["Rapporti cambio individuo migliore: ............", num2str(r_tot(:,:,end)(ind_best,1:nr))])
disp(["Regime di cambiata in ogni rapporto: ...........", num2str(r_tot(:,:,end)(ind_best,nr+3:end))])
disp(["Distanza b individuo migliore: .................", num2str(b_best*1000), "mm"])
disp(["Distribuzione dei pesi ant/post = ..............", num2str(b_best/p*100), "/", num2str((p-b_best)/p*100), "%"])
disp(["Altezza baricentro individuo migliore: .........", num2str(h_best*1000), "mm"])

## Trovo le caratteristiche dell'individuo migliore assoluto
t_best_ass = min(min(t_tot(:,:)));
ind_best_ass = find(t_tot(:,:)==t_best_ass)(1);
ind_col = ceil(ind_best_ass/np);
ind_row = ind_best_ass-floor(ind_best_ass/np)*np;
if ind_row == 0
  ind_row = np;
endif
disp("")
disp(["Individuo migliore assoluto nella generazione ", num2str(ind_col), " di ", num2str(columns(t_tot))])
disp(["Tempo percorrenza: .............................", num2str(t_best_ass), "s"])
disp(["Rapporti cambio: ...............................", num2str(r_tot(:,:,ind_col)(ind_row,1:nr))])
disp(["Regime di cambiata in ogni rapporto: ...........", num2str(r_tot(:,:,ind_col)(ind_row,nr+3:end))])
disp(["Distanza b: ....................................", num2str(r_tot(:,:,ind_col)(ind_row,nr+1)*1000), "mm"])
disp(["Distribuzione dei pesi ant/post = ..............", num2str(r_tot(:,:,ind_col)(ind_row,nr+1)/p*100), "/", num2str((p-r_tot(:,:,ind_col)(ind_row,nr+1))/p*100), "%"])
disp(["Altezza baricentro: ............................", num2str(r_tot(:,:,ind_col)(ind_row,nr+2)*1000), "mm"])

t_fin = time();
disp("");
disp("FINE CALCOLO OTTIMIZZAZIONE GENETICA");
disp(ctime(t_fin));
time_tot = t_fin-t_ini;
if time_tot >= 3600
  disp(["Tempo totale impiegato: ", num2str(floor(time_tot/3600)), "h ", num2str(floor((time_tot/3600-floor(time_tot/3600))*60)), "m ", num2str((((time_tot/3600-floor(time_tot/3600))*60)-floor((time_tot/3600-floor(time_tot/3600))*60))*60), "s"]);
elseif time_tot >= 60
  disp(["Tempo totale impiegato: ", num2str(floor(time_tot/60)), "m ", num2str((time_tot/60-floor(time_tot/60))*60), "s"]);
else
  disp(["Tempo totale impiegato: ", num2str(time_tot), "s"]);
endif
diary off;
